/***************************************************************

	Turbo C++ 1.01-specific implementation of Bywater BASIC
	Portions copyright (C) 2003 Steve Nickolas <steve@dosius.zzn.com>
	Portions copyright (C) 2003 Eric Auer <eric@coli.uni-sb.de>
	This program is licensed under the GNU GPL.

        *** Revision History for this file:
            2003.0227 patched up to 2.20, fixed bwx_putc() [\n] -uso.
            2003.0312 dummy functions for MS style graphics / sound -ea.

****************************************************************

        bwx_iqc.c       Environment-dependent implementation
                        of Bywater BASIC Interpreter
			for IBM PC and Compatibles
			using the Microsoft QuickC (tm) Compiler

                        Copyright (c) 1993, Ted A. Campbell
			Bywater Software

                        email: tcamp@delphi.com

        Copyright and Permissions Information:

        All U.S. and international rights are claimed by the author,
        Ted A. Campbell.

	This software is released under the terms of the GNU General
	Public License (GPL), which is distributed with this software
	in the file "COPYING".  The GPL specifies the terms under
	which users may copy and use the software in this distribution.

	A separate license is available for commercial distribution,
	for information on which you should contact the author.

***************************************************************/

#define BWB_EXT_VER "Release 3"

/*---------------------------------------------------------------*/
/* NOTE: Modifications marked "JBV" were made by Jon B. Volkoff, */
/* 11/1995 (eidetics@cerf.net).                                  */
/*                                                               */
/* Those additionally marked with "DD" were at the suggestion of */
/* Dale DePriest (daled@cadence.com).                            */
/*---------------------------------------------------------------*/
/* Modifications marked "USO" are 2003 by Steve Nickolas,        */
/* <steve@dosius.zzn.com>.                                       */
/* Marked "EA": 2003 by Eric Auer <eric@coli.uni-sb.de>.         */
/*---------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <bios.h>
#include <conio.h>
#include <ctype.h> /* tolower() -ea */

#define interrupt /* remove this non-ANSI keyword. breaks stuff -ea */
#include <dos.h>
/* Turbo C dos.h makes use non-ANSI keywords "interrupt": -ea */
/* used only in getvect, setvect -ea */

#include <signal.h>

#include "bwbasic.h"
#include "bwb_mes.h"

#include "llvid.h" /* -ea */

extern unsigned _stklen = 8192U; /* duplicate definition!? -ea */

extern int prn_col;
extern jmp_buf mark;
short oldfgd;
long oldbgd;
int reset_mode = FALSE;

bnumber lastX = 0;
bnumber lastY = 0;
int lastColor = 15;

static int iqc_setpos( void );

/***************************************************************

        FUNCTION:       main()

        DESCRIPTION:    As in any C program, main() is the basic
                        function from which the rest of the
                        program is called. Some environments,
			however, provide their own main() functions
			(Microsoft Windows (tm) is an example).
			In these cases, the following code will
			have to be included in the initialization
			function that is called by the environment.

***************************************************************/

void
main( int argc, char **argv )
   {
#if MS_CMDS
   struct text_info vc;
   /* short videomode; */

   /* Save original foreground, background, and text position. */

   gettextinfo( &vc );

   if ( vc.currmode != C80 )
      {
	 textmode( C80 );
      }
   else
      {
      reset_mode = FALSE;
      }

#endif       /* MS_CMDS */

   bwb_init( argc, argv );

#if INTERACTIVE
   setjmp( mark );
#endif

   /* now set the number of colors available */

   * var_findnval( co, co->array_pos ) = 16; /* (bnumber) vc.numcolors; */

   /* main program loop */

   while( !feof( stdin ) )		/* condition !feof( stdin ) added in v1.11 */
      {
      bwb_mainloop();
      }

   }

/***************************************************************

        FUNCTION:       bwx_signon()

        DESCRIPTION:

***************************************************************/

int
bwx_signon( void )
   {

   sprintf( bwb_ebuf, "\r%s %s\n", MES_SIGNON, VERSION );
   prn_xprintf( stdout, bwb_ebuf );
   sprintf( bwb_ebuf, "\r%s\n", MES_COPYRIGHT );
   prn_xprintf( stdout, bwb_ebuf );

   /* Perfectly legal ;) -uso. */
   prn_xprintf( stdout, "Enhanced DOS Version, " BWB_EXT_VER "\n" );

   prn_xprintf( stdout, "Copyright (c) 2003 Steve Nickolas and others\n");

#if PERMANENT_DEBUG
   sprintf( bwb_ebuf, "\r%s\r\n", "Debugging Mode" );
   prn_xprintf( stdout, bwb_ebuf );
#else
   sprintf( bwb_ebuf, "\r%s\r\n", MES_LANGUAGE );
   prn_xprintf( stdout, bwb_ebuf );
#endif

   return TRUE;

   }

/***************************************************************

        FUNCTION:       bwx_message()

        DESCRIPTION:

***************************************************************/

int
bwx_message( char *m )
   {

#if DEBUG
   cprintf( "<MES>" );
#endif

   cprintf( "%s", m );

   return TRUE;

   }

/***************************************************************

	FUNCTION:       bwx_putc()

	DESCRIPTION:

***************************************************************/

extern int
bwx_putc( char c )
   {
    if (c=='\n') cprintf("\r\n"); else putch (c);
    return TRUE;
   }

/***************************************************************

        FUNCTION:       bwx_error()

        DESCRIPTION:

***************************************************************/

int
bwx_errmes( char *m )
   {
   static char tbuf[ MAXSTRINGSIZE + 1 ];	/* this memory should be
						   permanent in case of memory
						   overrun errors */

   if (( prn_col != 1 ) && ( errfdevice == stderr ))
      {
      prn_xprintf( errfdevice, "\n" );
      }
   if ( CURTASK number == 0 )
      {
      sprintf( tbuf, "\r\n%s: %s\r\n", ERRD_HEADER, m );
      }
   else
      {
      sprintf( tbuf, "\r\n%s %d: %s\r\n", ERROR_HEADER, CURTASK number, m );
      }

#if INTENSIVE_DEBUG
   prn_xprintf( stderr, "<ERR>" );
#endif

   prn_xprintf( errfdevice, tbuf );

   return TRUE;

   }

/***************************************************************

        FUNCTION:       bwx_input()

	DESCRIPTION:    As implemented here, the input facility
			is a hybrid of _outtext output (which allows
			the color to be set) and standard output
			(which does not).  The reason is that I've
			found it helpful to use the DOS facility
			for text entry, with its backspace-delete
			and recognition of the SIGINT, depite the
			fact that its output goes to stdout.

***************************************************************/

int
bwx_input( char *prompt, char *buffer )
   {

#if INTENSIVE_DEBUG
   prn_xprintf( stdout, "<INP>" );
#endif

   prn_xprintf( stdout, prompt );

   fgets( buffer, MAXREADLINESIZE, stdin );
   prn_xprintf( stdout, "\n" );               /* let _outtext catch up */

   * prn_getcol( stdout ) = 1;			/* reset column */

   return TRUE;

   }

/***************************************************************

        FUNCTION:       bwx_terminate()

        DESCRIPTION:

***************************************************************/

void
bwx_terminate( void )
   {
#if MS_CMDS

   if ( reset_mode == TRUE )
      {

       textmode( C80 ); /* FIXME - should restore old mode */

      }

#endif

   exit( 0 );
   }

/***************************************************************

        FUNCTION:       itc_whine()

        DESCRIPTION:	Checks if an expression is of appropriate
        		type and prints an error message otherwise.
        		Returns TRUE on error. By EA.

***************************************************************/

int itc_whine( struct exp_ese *e,
  int wish, char *where )
{
   if (e->type != wish)
   {
#if PROG_ERRORS
     sprintf( bwb_ebuf, "in %s: Argument must be a %s", where,
       (wish == STRING) ? "string" : "number");
     bwb_error( bwb_ebuf );
#else
     if (wish == STRING)
        {
        bwb_error( err_argstr );
        }
     else
        {
        bwb_error( err_syntax );
        }
#endif
     return TRUE;
   }
   return FALSE;

  /* PS: when PROG_ERRORS is on, {exp,var}_get{n,s}val */
  /* return almost useful error messages anyway...     */
}

/***************************************************************

	FUNCTION:       bwx_shell()

	DESCRIPTION:

***************************************************************/

#if COMMAND_SHELL
extern int
bwx_shell( struct bwb_line *l )
   {
   static char *s_buffer;
   static int init = FALSE;
   static int position;

   /* get memory for temporary buffer if necessary */

   if ( init == FALSE )
      {
      init = TRUE;

      /* Revised to CALLOC pass-thru call by JBV */ /* ,0 added to aid compile -uso. */
      if ( ( s_buffer = CALLOC( MAXSTRINGSIZE + 1, sizeof( char ) ,0 )) == NULL )
	 {
	 bwb_error( err_getmem );
	 return FALSE;
	 }
      }

   /* get the first element and check for a line number */

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwx_shell(): line buffer is <%s>.", l->buffer );
   bwb_debug( bwb_ebuf );
#endif

   position = 0;
   adv_element( l->buffer, &position, s_buffer );
   if ( is_numconst( s_buffer ) != TRUE )                  /* not a line number */
      {

#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "in bwx_shell(): no line number, command <%s>.",
	 l->buffer );
      bwb_debug( bwb_ebuf );
#endif

      if ( system( l->buffer ) == 0 )
	 {
	 iqc_setpos();
	 return TRUE;
	 }
      else
	 {
	 iqc_setpos();
	 return FALSE;
	 }
      }

   else                                         /* advance past line number */
      {
      adv_ws( l->buffer, &position );           /* advance past whitespace */

#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "in bwx_shell(): line number, command <%s>.",
	 l->buffer );
      bwb_debug( bwb_ebuf );
#endif

      if ( system( &( l->buffer[ position ] ) ) == 0 )
	 {
	 iqc_setpos();
	 return TRUE;
	 }
      else
	 {
	 iqc_setpos();
	 return FALSE;
	 }
      }
   }
#endif

/***************************************************************

	FUNCTION:      iqc_setpos()

	DESCRIPTION:

***************************************************************/

static int
iqc_setpos( void )
   {
   union REGS ibm_registers;

   /* call the BDOS function 0x10 to read the current cursor position */

   ibm_registers.h.ah = 3;
   ibm_registers.h.bh = (unsigned char) 0/*_getvisualpage() -> FIXME*/;
   int86( 0x10, &ibm_registers, &ibm_registers );

   /* set text to this position */

   gotoxy( ibm_registers.h.dl, ibm_registers.h.dh );

   /* and move down one position */

   prn_xprintf( stdout, "\n" );

   return TRUE;
   }


#if COMMON_CMDS

/***************************************************************

        FUNCTION:       bwb_edit()

        DESCRIPTION:

***************************************************************/

struct bwb_line *
bwb_edit( struct bwb_line *l )
   {
   char tbuf[ MAXSTRINGSIZE + 1 ];
   char edname[ MAXSTRINGSIZE + 1 ];
   struct bwb_variable *ed;
   FILE *loadfile;

   ed = var_find( DEFVNAME_EDITOR );
   str_btoc( edname, var_getsval( ed ));

   sprintf( tbuf, "%s %s", edname, CURTASK progfile );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_edit(): command line <%s>", tbuf );
   bwb_debug( bwb_ebuf );
#else
   system( tbuf );
#endif

   /* open edited file for read */

   if ( ( loadfile = fopen( CURTASK progfile, "r" )) == NULL )
      {
      sprintf( bwb_ebuf, err_openfile, CURTASK progfile );
      bwb_error( bwb_ebuf );

      iqc_setpos();
      return bwb_zline( l );
      }

   /* clear current contents */

   bwb_new( l ); /* Relocated by JBV (bug found by DD) */

   /* and (re)load the file into memory */

   bwb_fload( loadfile );


   iqc_setpos();
   return bwb_zline( l );
   }

/***************************************************************

        FUNCTION:       bwb_renum()

	DESCRIPTION:    This function implements the BASIC RENUM
			command by shelling out to a default
			renumbering program called "renum".
			Added by JBV 10/95

	SYNTAX:		RENUM

***************************************************************/

#if ANSI_C
struct bwb_line *
bwb_renum( struct bwb_line *l )
#else
struct bwb_line *
bwb_renum( l )
   struct bwb_line *l;
#endif
   {
   char tbuf[ MAXSTRINGSIZE + 1 ];
   FILE *loadfile;

   sprintf( tbuf, "renum %s\0", CURTASK progfile );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_renum(): command line <%s>", tbuf );
   bwb_debug( bwb_ebuf );
#else
   system( tbuf );
#endif

   /* open edited file for read */

   if ( ( loadfile = fopen( CURTASK progfile, "r" )) == NULL )
      {
      sprintf( bwb_ebuf, err_openfile, CURTASK progfile );
      bwb_error( bwb_ebuf );

      iqc_setpos();
      return bwb_zline( l );
      }

   /* clear current contents */

   bwb_new( l ); /* Relocated by JBV (bug found by DD) */

   /* and (re)load the file into memory */

   bwb_fload( loadfile );


   iqc_setpos();
   return bwb_zline( l );
   }

/***************************************************************

        FUNCTION:       bwb_files()

        DESCRIPTION:

***************************************************************/

struct bwb_line *
bwb_files( struct bwb_line *l )
   {
   char tbuf[ MAXVARNAMESIZE + 1 ];
   char finame[ MAXVARNAMESIZE + 1 ];
   char argument[ MAXVARNAMESIZE + 1 ];
   struct bwb_variable *fi;
   struct exp_ese *e;

   fi = var_find( DEFVNAME_FILES );
   str_btoc( finame, var_getsval( fi ));

   /* get argument */

   adv_ws( l->buffer, &( l->position ));
   switch( l->buffer[ l->position ] )
      {
      case '\0':
      case '\r':
      case '\n':
         argument[ 0 ] = '\0';
         break;
      default:
         e = bwb_exp( l->buffer, FALSE, &( l->position ) );
         if ( e->type != STRING )
            {
            bwb_error( err_mismatch );
            return bwb_zline( l );
            }
         str_btoc( argument, exp_getsval( e ) );
         break;
      }

/* possible todo? replace with internal "files" command -uso. */

   sprintf( tbuf, "%s %s", finame, argument );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "in bwb_files(): command line <%s>", tbuf );
   bwb_debug( bwb_ebuf );
#else
   system( tbuf );
#endif

   iqc_setpos();
   return bwb_zline( l );

   }

#endif					/* COMMON_CMDS */

#if INTERACTIVE

/***************************************************************

        FUNCTION:       fnc_inkey()

        DESCRIPTION:    This C function implements the BASIC INKEY$
        		function.  It is implementation-specific.

***************************************************************/

/* Patched with unique_id -uso. */

/* Don't produce warning for unique_id */
#pragma argsused

extern struct bwb_variable *
fnc_inkey( int argc, struct bwb_variable *argv, int unique_id )
   {
   static struct bwb_variable nvar;
   char tbuf[ MAXSTRINGSIZE + 1 ];
   static int init = FALSE;

   if ((argv==NULL) && (unique_id==42)) unique_id++; /* no unused args -ea */

   /* initialize the variable if necessary */

   if ( init == FALSE )
      {
      init = TRUE;
      var_make( &nvar, STRING /*-uso. , "fnc_inkey"*/ );
      }

   /* check arguments */

#if PROG_ERRORS
   if ( argc > 0 )
      {
      sprintf( bwb_ebuf, "Too many arguments to function INKEY$()" );
      bwb_error( bwb_ebuf );
      return &nvar;
      }

#else
   if ( fnc_checkargs( argc, argv, 0, 0 ) == FALSE )
      {
      return NULL;
      }
#endif

   /* body of the INKEY$ function */

   if ( bioskey( 1 ) == 0 )
      {
      tbuf[ 0 ] = '\0';
      }
   else
      {
      tbuf[ 0 ] = (char) bioskey( 0 );
      tbuf[ 1 ] = '\0';
      }

   /* assign value to nvar variable */

   str_ctob( var_findsval( &nvar, nvar.array_pos ), tbuf );

   /* return value contained in nvar */

   return &nvar;

   }

#endif				/* INTERACTIVE */

#if MS_CMDS
/* CLS moved -uso. */

/***************************************************************

        FUNCTION:       bwb_locate()

        DESCRIPTION:    This C function implements the BASIC LOCATE
        		command.  It is implementation-specific.

***************************************************************/

extern struct bwb_line *
bwb_locate( struct bwb_line *l )
   {
   struct exp_ese *e;
   int row, column;

   /* get first argument */
   e = bwb_exp( l->buffer, FALSE, &( l->position ));
   if (itc_whine(e, NUMBER, "bwb_line"))
     return bwb_zline( l );
   row = (int)((bnumber) exp_getnval( e )); /* -ea */

   /* advance past comma */
   adv_ws( l->buffer, &( l->position ));
   if ( l->buffer[ l->position ] != ',' )
      {
      bwb_error( err_syntax );
      return bwb_zline( l );
      }
   ++( l->position );

   /* get second argument */
   e = bwb_exp( l->buffer, FALSE, &( l->position ));
   if (itc_whine(e, NUMBER, "bwb_line"))
     return bwb_zline( l );
   column = (int) exp_getnval( e );

   /* position the cursor */
   gotoxy( column, row );

   return bwb_zline( l );
   }

/***************************************************************

	FUNCTION:       bwb_color()

	DESCRIPTION:    This C function implements the BASIC COLOR
			command.  It is implementation-specific.

	SYNTAX:		COLOR foreground [, background]

***************************************************************/

extern struct bwb_line *
bwb_color( struct bwb_line *l )
   {
   struct exp_ese *e;
   int fgcolor, bgcolor = -1;

   /* get first argument */

   e = bwb_exp( l->buffer, FALSE, &( l->position ));

   if (itc_whine(e, NUMBER, "bwb_line"))
     return bwb_zline( l );
     
   fgcolor = (int) exp_getnval( e );

   textcolor( (short) fgcolor );

   /* advance past comma */

   adv_ws( l->buffer, &( l->position ));
   if ( l->buffer[ l->position ] != ',' )
      {
#if INTENSIVE_DEBUG
      sprintf( bwb_ebuf, "Set foreground color to: %d",
        fgcolor);
      bwb_debug( bwb_ebuf );
#endif
      return bwb_zline( l );
      }
   ++( l->position );

   /* get second argument */
   e = bwb_exp( l->buffer, FALSE, &( l->position ));
   if (itc_whine(e, NUMBER, "bwb_line"))
      return bwb_zline( l );
   bgcolor = (int) exp_getnval( e );

   /* set the background color */
   textbackground( /* (long) why long??? -ea */ bgcolor );

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "Set colors to foreground: %d background: %d\r\n",
     fgcolor, bgcolor );
   bwb_debug( bwb_ebuf );
#endif

   return bwb_zline( l );
   }

/***************************************************************

	FUNCTION:       bwb_draw()

	DESCRIPTION:    This C function implements the BASIC DRAW
			command.  It is implementation-specific.
			
	SYNTAX:		DRAW command-string

***************************************************************/

extern struct bwb_line *
bwb_draw( struct bwb_line *l )
   {
   struct exp_ese *e;
   char * cmd = NULL;

   /* get first argument */

   e = bwb_exp( l->buffer, FALSE, &( l->position ));

   if (itc_whine(e, STRING, "bwb_draw"))
     return bwb_zline( l );

   /* cmd = exp_getsval( e ); illegal conversion -ea */
   str_btoc( cmd , exp_getsval( e ) ); /* -ea */

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "DRAW command received: %s", cmd );
   bwb_debug( bwb_ebuf );
#endif

   /* ***************************************************** */
   /* TODO: parse the cmd string and do the turtle graphics */
   /* commands that the cmd string contains.                */
   /* TODO: this may have to use WINDOW data.               */
   /* ***************************************************** */
   
   fprintf(stderr, "DRAW: dummy (%s)\r\n", cmd);

   return bwb_zline( l );
   }

/***************************************************************

	FUNCTION:       bwb_play()

	DESCRIPTION:    This C function implements the BASIC PLAY
			command.  It is implementation-specific.
			
	SYNTAX:		PLAY command-string

***************************************************************/

extern struct bwb_line *
bwb_play( struct bwb_line *l )
   {
   struct exp_ese *e;
   char * cmd = NULL;

   /* get first argument */

   e = bwb_exp( l->buffer, FALSE, &( l->position ));

   if (itc_whine(e, STRING, "bwb_draw"))
     return bwb_zline( l );

   /* cmd = exp_getsval( e ); illegal conversion -ea */
   str_btoc( cmd , exp_getsval( e ) ); /* -ea */

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "DRAW command received: %s", cmd );
   bwb_debug( bwb_ebuf );
#endif

   /* ***************************************************** */
   /* TODO: parse the cmd string and do the sequence of     */
   /* sound commands that the cmd string requests.          */
   /* TODO: this may include background playing.            */
   /* ***************************************************** */
   
   fprintf(stderr, "PLAY: dummy (%s)\r\n", cmd);

   return bwb_zline( l );
   }

/***************************************************************

        FUNCTION:       fnc_play()

        DESCRIPTION:    Returns the remaining number of notes
                        to be played by background mode PLAY.
                        Returns 0 as we have no background PLAY.

        SYNTAX:		PLAY( 0 )

***************************************************************/

struct bwb_variable *
fnc_play( int argc, struct bwb_variable *argv, int unique_id )
   {
   static struct bwb_variable nvar;
   static int init = FALSE;

   if ((!argc) && (argc==NULL) && (!unique_id)) /* dummy */
     unique_id++; /* avoid unused args -ea */
     
   /* initialize the variable if necessary */

   if ( init == FALSE )
      {
      init = TRUE;
      var_make( &nvar, NUMBER );
      }

   if ( fnc_checkargs( argc, argv, 1, 1 ) == FALSE )
      {
#if PROG_ERRORS
      bwb_error( "Must give one (dummy) argument to PLAY()" );
#endif
      return NULL;
      }

   * var_findnval( &nvar, nvar.array_pos ) = (bnumber) 0;

   return &nvar;
   }

/***************************************************************

        FUNCTION:       fnc_point()

        DESCRIPTION:    Returns the color of the point at the
                        given screen coordinates. Special
                        one argument syntax not yet implemented.

        SYNTAX:		POINT( x, y )   (get a pixel)
			POINT( number ) (not implemented!)

***************************************************************/

struct bwb_variable *
fnc_point( int argc, struct bwb_variable *argv, int unique_id )
   {
   static struct bwb_variable nvar;
   static int init = FALSE;

   if ((!argc) && (argc==NULL) && (!unique_id)) /* dummy */
     unique_id++; /* avoid unused args -ea */

   /* initialize the variable if necessary */

   if ( init == FALSE )
      {
      init = TRUE;
      var_make( &nvar, NUMBER );
      }

   if ( fnc_checkargs( argc, argv, 2, 2 ) == FALSE )
      {
#if PROG_ERRORS
      sprintf(bwb_ebuf,
        "Only POINT(x,y) allowed yet, but %d arguments found" , argc );
      bwb_error( bwb_ebuf ); /* -ea */
#endif
      return NULL;
      }

   /* TODO: scale if WINDOW is used */

   * var_findnval( &nvar, nvar.array_pos ) = (bnumber)
     llvid_point( var_getnval( &( argv[ 0 ] ) ),
                  var_getnval( &( argv[ 1 ] ) ) );

   return &nvar;
   }

/***************************************************************

        FUNCTION:       fnc_screen()

        DESCRIPTION:    Returns the character found at the
                        given screen coordinates. Special
                        N argument syntax not yet done.

        SYNTAX:		SCREEN( x, y )    (get a character)
			SCREEN( x, y, z ) (get attribute if z is 1)

***************************************************************/

struct bwb_variable *
fnc_screen( int argc, struct bwb_variable *argv, int unique_id )
   {
   static struct bwb_variable nvar;
   static int init = FALSE;
   int mode = 0;

   if ((!argc) && (argc==NULL) && (!unique_id)) /* dummy */
     unique_id++; /* avoid unused args -ea */

   /* initialize the variable if necessary */

   if ( init == FALSE )
      {
      init = TRUE;
      var_make( &nvar, NUMBER );
      }

   if ( fnc_checkargs( argc, argv, 2, 3 ) == FALSE )
      {
#if PROG_ERRORS
      sprintf(bwb_ebuf,
        "Only SCREEN(x,y[,mode]) allowed, but %d arguments found", argc);
      bwb_error( bwb_ebuf ); /* -ea */
#endif
      return NULL;
      }

   if (argc == 3)
     mode = var_getnval( &( argv[ 2 ] ) );

   if ((mode < 0) || (mode > 1))
     {
     bwb_error( err_valoorange );
     return NULL;
     }

   * var_findnval( &nvar, nvar.array_pos ) = (bnumber)
     llvid_screen_func( var_getnval( &( argv[ 0 ] ) ),
                        var_getnval( &( argv[ 1 ] ) ), mode );

   return &nvar;
   }

/***************************************************************

        FUNCTION:       bwb_circle()

        DESCRIPTION:    This C function implements the BASIC CIRCLE
        		command.  It is implementation-specific.

	SYNTAX:		CIRCLE [STEP] (x,y), radius
			[,color [,[ang1],[ang2][,aspect]]]
			(angle / aspect stuff not implemented yet)

***************************************************************/

extern struct bwb_line *
bwb_circle( struct bwb_line *l )
   {
   struct exp_ese *e;
   int xi, yi, radiusi;
   bnumber x, y, radius;
   int color = lastColor;
   int havemore = TRUE;
   bnumber angle1 = 0;
   bnumber angle2 = 360; /* ??? */
   bnumber aspect = 1;

   /* relative versus absolute coordinates */
   if ( tolower(l->buffer[ l->position]) == 's' &&
        tolower(l->buffer[ l->position + 1]) == 't' &&
        tolower(l->buffer[ l->position + 2]) == 'e' &&
        tolower(l->buffer[ l->position + 3]) == 'p'
      ) /* string is not char *: using && not strcasecmp -ea */
      {
      (l->position) += 4;
      adv_ws( l->buffer, &( l->position ));
      x = lastX;
      y = lastY;
      }
   else
      {
      x = 0; /* or WINDOW origin? */
      y = 0;
      }

   /* advance past parenthesis */
   adv_ws( l->buffer, &( l->position ));
   if ( l->buffer[ l->position ] != '(' )
      {
      bwb_error( err_syntax );
      return bwb_zline( l );
      }
   ++( l->position );

   /* get x argument, possibly relative */
   e = bwb_exp( l->buffer, FALSE, &( l->position ));
   if (itc_whine(e, NUMBER, "bwb_circle"))
     return bwb_zline( l );
   x = x + exp_getnval( e ); /* unscaled */

   /* advance past comma */
   adv_ws( l->buffer, &( l->position ));
   if ( l->buffer[ l->position ] != ',' )
      {
      bwb_error( err_syntax );
      return bwb_zline( l );
      }
   ++( l->position );

   /* get y argument, possibly relative */
   e = bwb_exp( l->buffer, FALSE, &( l->position ));
   if (itc_whine(e, NUMBER, "bwb_circle"))
     return bwb_zline( l );
   y = y + exp_getnval( e ); /* unscaled */

   /* advance past parenthesis */
   adv_ws( l->buffer, &( l->position ));
   if ( l->buffer[ l->position ] != ')' )
      {
      bwb_error( err_syntax );
      return bwb_zline( l );
      }
   ++( l->position );

   /* get radius argument */
   e = bwb_exp( l->buffer, FALSE, &( l->position ));
   if (itc_whine(e, NUMBER, "bwb_circle"))
      return bwb_zline( l );
   radius = exp_getnval( e ); /* unscaled */

   /* advance past comma */
   adv_ws( l->buffer, &( l->position ));
   if ( l->buffer[ l->position ] != ',' )
      {
      havemore = FALSE;
      color = lastColor;
      }

   if (havemore)
      {
      ++( l->position );
      /* get optional color argument */
      e = bwb_exp( l->buffer, FALSE, &( l->position ));
      if (itc_whine(e, NUMBER, "bwb_circle"))
         return bwb_zline( l );
      color = (int)exp_getnval( e );
      /* advance past comma */
      adv_ws( l->buffer, &( l->position ));
      if ( l->buffer[ l->position ] != ',' )
         {
         havemore = FALSE;
         }
      }

   if (havemore)
      {
      ++( l->position );
      /* get optional first angle argument */
      if ( l->buffer[ l->position ] != ',')
         {
         e = bwb_exp( l->buffer, FALSE, &( l->position ));
         if (itc_whine(e, NUMBER, "bwb_circle"))
            return bwb_zline( l );
         angle1 = exp_getnval( e );
         }
      /* advance past comma */
      adv_ws( l->buffer, &( l->position ));
      if ( l->buffer[ l->position ] != ',' )
         {
         havemore = FALSE;
         }
      }

   if (havemore)
      {
      ++( l->position );
      /* get optional second angle argument */
      if ( l->buffer[ l->position ] != ',')
         {
         e = bwb_exp( l->buffer, FALSE, &( l->position ));
         if (itc_whine(e, NUMBER, "bwb_circle"))
            return bwb_zline( l );
         angle2 = exp_getnval( e );
         }
      /* advance past comma */
      adv_ws( l->buffer, &( l->position ));
      if ( l->buffer[ l->position ] != ',' )
         {
         havemore = FALSE;
         }
      }

   if (havemore)
      {
      ++( l->position );
      /* get optional aspect argument */
      e = bwb_exp( l->buffer, FALSE, &( l->position ));
      if (itc_whine(e, NUMBER, "bwb_circle"))
         return bwb_zline( l );
      aspect = exp_getnval( e );
      /* } -ea */
      /* advance past comma */
      adv_ws( l->buffer, &( l->position ));
      if ( l->buffer[ l->position ] != ',' )
         {
         havemore = FALSE;
         }
      }

   if (havemore)
      {
#if PROG_ERRORS
      bwb_error( "Too many arguments to CIRCLE" );
#else
      bwb_error( err_syntax );
#endif
      }

   /* TODO: -scale- if WINDOW is used */
   /* (WINDOW -origin- already prepared above!?!?) */
   xi = (int) x;
   yi = (int) y;
   radiusi = (int) radius;

#if INTENSIVE_DEBUG
   sprintf( bwb_ebuf, "CIRCLE: x=%d, y=%d, r=%f, color=%d, phi1=%f, phi2=%f, aspect=%f",
     xi, yi, radiusi, color, angle1, angle2, aspect );
   bwb_debug( bwb_ebuf );
#endif

   /* ******** TODO: llvid_circle is not implemented yet ******** */
   sprintf( bwb_ebuf, "CIRCLE: x=%d, y=%d, r=%f, color=%d, phi1=%f, phi2=%f, aspect=%f",
     xi, yi, radiusi, color, angle1, angle2, aspect );
   bwb_debug( bwb_ebuf );

   /* draw that circle */
   /* Use my Bresenham circles code snippet plus a wrapper that uses  */
   /* line segments instead if angles are used. Thanks. -ea           */
   /* llvid_circle( xi, yi, radiusi, color, angle1, angle2, aspect ); */

   fprintf( stderr, "DUMMY CIRCLE: x=%d, y=%d, r=%f, color=%d, phi1=%f, phi2=%f, aspect=%f",
     xi, yi, radiusi, color, angle1, angle2, aspect );

   lastX = x;
   lastY = y;

   return bwb_zline( l );
   }

/* ******************************************************* */
/* TODO: bwb_pset bwb_paint bwb_line bwb_screen bwb_window */
/* TODO: bwb_sound                                         */
/* TODO: "undummify" DRAW and PLAY above.                  */
/* TODO: add line (llvid_line should allow  fill and       */
/*       rectangle mode) and circle primitives to llvid.   */
/* TODO: processing of WINDOW settings                     */
/* TODO: make code shorter by having ONE parser for all    */
/*       places where [STEP] (x, y) is allowed ("coord").  */
/*       This could also do WINDOW scaling for coords.     */
/* ******************************************************* */

extern struct bwb_line *
bwb_pset( struct bwb_line *l ) /* paint a pixel */
   {
   /* coord [,color] */
   fprintf( stderr, "DUMMY: PSET" );
   return bwb_zline( l );
   }
extern struct bwb_line *
bwb_paint( struct bwb_line *l ) /* fill starting at point */
   {
   /* coord, color, bgcolor */
   fprintf( stderr, "DUMMY: PAINT" );
   return bwb_zline( l );
   }
extern struct bwb_line *
bwb_linecmd( struct bwb_line *l ) /* draw a line or rectangle */
   {
   /* [coord] - coord [,color [,[B[F]]]] */
   fprintf( stderr, "DUMMY: LINE" );
   return bwb_zline( l );
   }
extern struct bwb_line *
bwb_screen( struct bwb_line *l ) /* set a video mode */
   {
   /* mode [, ?, apage, vpage]  ??? */
   fprintf( stderr, "DUMMY: SCREEN" );
   return bwb_zline( l );
   }
extern struct bwb_line *
bwb_window( struct bwb_line *l ) /* configure scaling */
   {
   /* coord, coord */
   fprintf( stderr, "DUMMY: WINDOW" );
   return bwb_zline( l );
   }
extern struct bwb_line *
bwb_sound( struct bwb_line *l ) /* play a sound with the beeper */
   {
   /* frequency, duration-in-ticks */
   /* ticks are DOS system timer ticks, 1 / 18.2 seconds each */
   fprintf( stderr, "DUMMY: SOUND" );
   return bwb_zline( l );
   }

#endif				/* MS_CMDS */

